@@ -1,6 +1,10 @@
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE DataKinds #-}
+{-# LANGUAGE DeriveAnyClass #-}
+{-# LANGUAGE DeriveGeneric #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE MagicHash #-}
 {-# LANGUAGE MonoLocalBinds #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
@@ -10,6 +14,7 @@
 {-# LANGUAGE UndecidableInstances #-}
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
+import qualified Data.HashMap.Strict as Map
 import           Data.Name
 import           Data.Parameterized.Context ( pattern Empty, pattern (:>) )
 import           Data.Proxy ( Proxy(Proxy) )
@@ -17,6 +22,7 @@ import           Data.String ( IsString(fromString) )
 import           Data.Text ( Text )
 import qualified Data.Text as T
 import           GHC.Exts ( proxy#, IsList(fromList, toList) )
+import           GHC.Generics ( Generic )
 import qualified Prettyprinter as PP
 import           Text.Sayable
 
@@ -26,6 +32,12 @@ import           Test.Tasty.Checklist
 import           Test.Tasty.Hspec
 import           Test.Tasty.Runners.AntXML
 
+#ifdef VERSION_aeson
+import           Data.Aeson
+import           Data.ByteString.Lazy ( ByteString )
+import           Data.Name.JSON
+#endif
+
 
 main :: IO ()
 main = tests >>= defaultMainWithIngredients (antXMLRunner : defaultIngredients)
@@ -42,6 +54,9 @@ tests = testGroup "Named" <$> sequence
         , testSomeNames
         , testHasName
         , testUtilities
+#ifdef VERSION_aeson
+        , testJSON
+#endif
         ]
 
 
@@ -296,6 +311,8 @@ testConversions = testSpec "Named Conversions" $ do
        :> Val "matches implicit construction" id "list of text"
        :> Val "as sayable" (sez @"test") "CR40-2 'list of text'"
        :> Val "as extracted text" nameText "list of text"
+       :> Val "nameOf" (\n -> nameOf n proxy#) "CR40-2"
+       :> Val "style proxy" styleProxy (Proxy :: Proxy UTF8)
       )
 
     it "CR41 UTF8 explicit conversion" $
@@ -306,6 +323,8 @@ testConversions = testSpec "Named Conversions" $ do
        :> Val "matches implicit construction" id "mjtu!pg!ufyu"
        :> Val "as sayable" (sez @"test") "CR41-3 'mjtu!pg!ufyu'"
        :> Val "as extracted text" nameText "mjtu!pg!ufyu"
+       :> Val "nameOf" (\n -> nameOf n proxy#) "CR41-3"
+       :> Val "style proxy" styleProxy (Proxy :: Proxy UTF8)
       )
 
     it "CR42 CaseInsensitive conversion" $
@@ -317,6 +336,8 @@ testConversions = testSpec "Named Conversions" $ do
        :> Val "matches implicit construction" id "bit of text"
        :> Val "as sayable" (sez @"test") "CR42 new «bit of text»"
        :> Val "as extracted text" nameText "bit of text"
+       :> Val "nameOf" (\n -> nameOf n proxy#) "CR42 new"
+       :> Val "style proxy" styleProxy (Proxy :: Proxy CaseInsensitive)
       )
 
     it "CR43 Secure conversion" $
@@ -329,6 +350,8 @@ testConversions = testSpec "Named Conversions" $ do
        :> Val "as sayable" (sez @"test") "CR43 again 'hi#######xt'"
        :> Val "as extracted text" nameText "hi#######xt"
        :> Val "security bypass extraction" secureNameBypass ("hidden text" :: Text)
+       :> Val "nameOf" (\n -> nameOf n proxy#) "CR43 again"
+       :> Val "style proxy" styleProxy (Proxy :: Proxy Secure)
       )
 
   describe "Named style conversions" $ do
@@ -343,6 +366,8 @@ testConversions = testSpec "Named Conversions" $ do
        :> Val "matches implicit construction" id "some text"
        :> Val "as sayable" (sez @"test") "CR44 «some text»"
        :> Val "as extracted text" nameText "some text"
+       :> Val "nameOf" (\n -> nameOf n proxy#) "CR44"
+       :> Val "style proxy" styleProxy (Proxy :: Proxy CaseInsensitive)
       )
 
     it "CR45 UTF8->Secure default conversion" $
@@ -353,6 +378,8 @@ testConversions = testSpec "Named Conversions" $ do
        :> Val "matches implicit construction" id "Some TEXT"
        :> Val "as sayable" (sez @"test") "CR45 'So#####XT'"
        :> Val "as extracted text" nameText "So#####XT"
+       :> Val "nameOf" (\n -> nameOf n proxy#) "CR45"
+       :> Val "style proxy" styleProxy (Proxy :: Proxy Secure)
       )
 
     it "CR46 CaseInsensitive->UTF8 default conversion" $
@@ -364,6 +391,8 @@ testConversions = testSpec "Named Conversions" $ do
        :> Val "matches implicit construction" id "some text"
        :> Val "as sayable" (sez @"test") "CR46 'some text'"
        :> Val "as extracted text" nameText "some text"
+       :> Val "nameOf" (\n -> nameOf n proxy#) "CR46"
+       :> Val "style proxy" styleProxy (Proxy :: Proxy UTF8)
       )
 
     it "CR47 Secure->UTF8 default conversion" $
@@ -375,6 +404,8 @@ testConversions = testSpec "Named Conversions" $ do
        :> Val "matches implicit construction" id "So#####XT"
        :> Val "as sayable" (sez @"test") "CR47 'So#####XT'"
        :> Val "as extracted text" nameText "So#####XT"
+       :> Val "nameOf" (\n -> nameOf n proxy#) "CR47"
+       :> Val "style proxy" styleProxy (Proxy :: Proxy UTF8)
       )
 
     it "CR48 Secure->UTF8 implicit bypass conversion" $
@@ -386,6 +417,8 @@ testConversions = testSpec "Named Conversions" $ do
        :> Val "matches implicit construction" id "Some TEXT"
        :> Val "as sayable" (sez @"test") "CR48 'Some TEXT'"
        :> Val "as extracted text" nameText "Some TEXT"
+       :> Val "nameOf" (\n -> nameOf n proxy#) "CR48"
+       :> Val "style proxy" styleProxy (Proxy :: Proxy UTF8)
       )
 
 instance ConvertName UTF8 "CR40" "CR40-2"
@@ -395,7 +428,6 @@ instance ConvertName CaseInsensitive "CR42" "CR42 new"
 instance ConvertName Secure "CR43" "CR43 again" where
   convertName = fromText . secureNameBypass
 
-instance ConvertNameStyle UTF8 CaseInsensitive "CR44"
 instance ConvertNameStyle UTF8 Secure "CR45"
 instance ConvertNameStyle CaseInsensitive UTF8 "CR46"
 instance ConvertNameStyle Secure UTF8 "CR47"
@@ -473,17 +505,21 @@ testHasName = testSpec "HasName" $ do
   describe "HasName Foo" $ do
 
     it "CR70 can extract UTF8 myName from Foo" $
-      myName (Foo "bar" "baz" "quux") `shouldBe` ("baz" :: Name "principle")
+      myName (Foo "bar" "baz" (Map.fromList [("one","quux"), ("two","brox")]))
+      `shouldBe` ("baz" :: Name "principle")
 
     it "CR71 can extract secure myName from Bar" $
       myName (Bar "baz" "quux") `shouldBe` ("quux" :: Named Secure "second")
 
-data Foo = Foo (Name "prefix") (Name "principle") (Name "alt")
+data Foo = Foo (Name "prefix") (Name "principle")
+               (Map.HashMap (Name "alt-key") (Name "alt"))
+         deriving (Eq, Generic, Show)
 instance HasName Foo UTF8 "principle" where myName (Foo _ p _) = p
 
 data Bar = Bar (Name "first") (Named Secure "second")
 instance HasName Bar Secure "second" where myName (Bar _ s) = s
 
+instance TestShow Foo
 
 testUtilities :: IO TestTree
 testUtilities = testSpec "Named utilities" $ do
@@ -517,3 +553,173 @@ testUtilities = testSpec "Named utilities" $ do
 
     it "CR88 can check a non-null Secure named" $
       nullName ("Not empty" :: Named Secure "CR88") `shouldBe` False
+
+----------------------------------------------------------------------
+-- Data.Name.JSON
+
+#ifdef VERSION_aeson
+
+instance TestShow (Proxy JSONStyle) where testShow _ = "Proxy :: \"JSONStyle\""
+
+testJSON :: IO TestTree
+testJSON = testSpec "Named JSON style" $ do
+  describe "Named JSON" $ do
+
+    it "CR90 can create a JSON Named from Text" $
+      withChecklist "CR90" $
+      fromText @(Named JSONStyle "CR90") ("test JSON text" :: Text)
+      `checkValues`
+      (Empty
+       :> Val "overloaded equivalent text value" id "test JSON text"
+       :> Got "case sensitive text value" (/= "Test json Text")
+       :> Val "name parameter value" (\n -> nameOf n proxy#) "CR90"
+       :> Val "style proxy" styleProxy (Proxy :: Proxy JSONStyle)
+       :> Val "extracted text" nameText ("test JSON text" :: Text)
+      )
+
+    it "CR91 render UTF8 via Sayable" $
+      withChecklist "CR91" $
+      ("test JSON thing" :: Named JSONStyle "CR91")
+      `checkValues`
+      (Empty
+       :> Val "as sayable" (sez @"test") "CR91 'test JSON thing'"
+       :> Val "as sayable info" (sez @"info") "test JSON thing"
+      )
+
+    it "CR92 render UTF8 via Prettyprinter" $
+      withChecklist "CR92" $
+      fromText @(Named JSONStyle "CR92") ("test of JSON text" :: Text)
+      `checkValues`
+      (Empty
+       :> Val "as pretty" (show . PP.pretty) "CR92 'test of JSON text'"
+      )
+
+    it "CR93 render UTF8 via Show" $
+      withChecklist "CR93" $
+      fromText @(Named JSONStyle "CR93") ("test of text" :: Text)
+      `checkValues`
+      (Empty
+       :> Val "as show" show "CR93 'test of text'"
+      )
+
+    it "CR94 JSON semigroup" $
+      withChecklist "CR94" $
+      (fromText @(Named JSONStyle "CR94") ("more test JSON text" :: Text) <> "!!")
+      `checkValues`
+      (Empty
+       :> Val "raw value" id "more test JSON text!!"
+       :> Val "as sayable" (sez @"test") "CR94 'more test JSON text!!'"
+      )
+
+    it "CR95 JSON default conversion" $
+      withChecklist "CR95" $
+      (convertName (fromText "some JSON text" :: Named JSONStyle "CR95")
+       :: Named JSONStyle "CR95 2")
+      `checkValues`
+      (Empty
+       :> Val "matches implicit construction" id "some JSON text"
+       :> Val "as sayable" (sez @"test") "CR95 2 'some JSON text'"
+       :> Val "as extracted text" nameText "some JSON text"
+       :> Val "nameOf" (\n -> nameOf n proxy#) "CR95 2"
+       :> Val "style proxy" styleProxy (Proxy :: Proxy JSONStyle)
+      )
+
+    it "CR96 UTF8->JSON default conversion" $
+      withChecklist "CR96" $
+      (convertStyle (fromText "Some JSON-able TEXT" :: Name "CR96")
+        :: Named JSONStyle "CR96")
+      `checkValues`
+      (Empty
+       :> Val "matches implicit construction" id "Some JSON-able TEXT"
+       :> Val "as sayable" (sez @"test") "CR96 'Some JSON-able TEXT'"
+       :> Val "as extracted text" nameText "Some JSON-able TEXT"
+      )
+
+    it "CR97 JSON->UTF8 default conversion" $
+      withChecklist "CR97" $
+      (convertStyle (fromText "Some un-JSON-able TEXT" :: Named JSONStyle "CR97")
+        :: Name "CR97")
+      `checkValues`
+      (Empty
+       :> Val "matches implicit construction" id "Some un-JSON-able TEXT"
+       :> Val "as sayable" (sez @"test") "CR97 'Some un-JSON-able TEXT'"
+       :> Val "as extracted text" nameText "Some un-JSON-able TEXT"
+      )
+
+    it "CR98 can get a JSON length" $
+      nameLength ("Length of JSON TEXT" :: Named JSONStyle "CR98") `shouldBe` 19
+
+    it "CR99 can check a null JSON Named" $
+      nullName ("" :: Named JSONStyle "CR99") `shouldBe` True
+
+    it "CR100 can check a non-null JSON Named" $
+      nullName ("*" :: Named JSONStyle "CR99") `shouldBe` False
+
+    it "CR101 Named toJSON is valid" $
+      withChecklist "CR101" $
+      let obj = convertStyle ("Regular TEXT!" :: Name "CR101")
+                :: Named JSONStyle "CR101"
+      in encode (toJSON obj)
+         `checkValues`
+         (Empty
+          :> Val "encoded" id "\"Regular TEXT!\""
+          :> Val "decoded" decode (Just obj)
+         )
+
+    it "CR102 toJSON of structure containing Named is valid" $
+      withChecklist "CR102" $
+      let obj = Foo "start" "your engines"
+                (Map.fromList [ ("one","ready")
+                              , ("two", "set")
+                              , ("three", "go")
+                              ])
+      in encode (toJSON obj)
+         `checkValues`
+         (Empty
+          :> Val "encoded" id
+          "[\"start\",\"your engines\",{\"one\":\"ready\",\"three\":\"go\",\"two\":\"set\"}]"
+          :> Val "decoded" decode (Just obj)
+         )
+
+    it "CR103 toJSON of record containing Named is valid" $
+      withChecklist "CR103" $
+      let obj = Info "John Henry" "railroad worker" "hammer master"
+      in encode (toJSON obj)
+         `checkValues`
+         (Empty
+          :> Val "encoded" id
+          "{\"desc\":\"hammer master\",\"name\":\"John Henry\",\"title\":\"railroad worker\"}"
+          :> Val "decoded" decode (Just obj)
+         )
+
+    it "CR104 JSON of record containing CaseInsensitive Named is not necessarily round-robin" $
+      withChecklist "CR104" $
+      let obj = Info "John Henry" "Railroad Worker" "Hammer Master"
+      in encode (toJSON obj)
+         `checkValues`
+         (Empty
+          :> Val "encoded" id
+          "{\"desc\":\"Hammer Master\",\"name\":\"John Henry\",\"title\":\"railroad worker\"}"
+          :> Val "decoded" decode (Just $ obj { title = fromText $ T.toLower $ nameText $ title obj })
+         )
+
+data Info = Info { name :: Name "name"
+                 , title :: Named CaseInsensitive "title"
+                 , desc :: Name "description"
+                 }
+  deriving (Eq, Generic, Show)
+instance ToJSON Info
+instance FromJSON Info
+
+instance ToJSON Foo
+instance FromJSON Foo
+
+instance ConvertName JSONStyle "CR95" "CR95 2"
+
+instance TestShow ByteString
+instance TestShow a => TestShow (Maybe a) where
+  testShow = \case
+    Nothing -> "NOTHING"
+    Just a -> "JUST " <> testShow a
+instance TestShow Info
+#endif
